package main

import (
	"bufio"
	"context"
	"flag"
	"fmt"
	"io"
	"log"
	"os"
	"regexp"

	ippb "git.autistici.org/ai3/tools/iprep/proto"
	"git.autistici.org/ai3/tools/iprep/submission"
	"github.com/google/subcommands"
)

type tailCommand struct {
	serverAddr  string
	patternFile string
}

type pattern struct {
	rx        *regexp.Regexp
	eventType string
}

var patternRx = regexp.MustCompile(`^\s*/(.*)/\s+(\w+)\s*$`)

func parsePatterns(r io.Reader, filename string) ([]pattern, error) {
	var out []pattern
	scanner := bufio.NewScanner(r)
	var lineno int
	for scanner.Scan() {
		lineno++
		m := patternRx.FindSubmatch(scanner.Bytes())
		if len(m) < 1 {
			continue
		}
		rx, err := regexp.Compile(string(m[1]))
		if err != nil {
			return nil, fmt.Errorf("syntax error, %s line %d: %v", filename, lineno, err)
		}
		out = append(out, pattern{
			rx:        rx,
			eventType: string(m[2]),
		})
	}
	return out, scanner.Err()
}

func (c *tailCommand) Name() string     { return "tail" }
func (c *tailCommand) Synopsis() string { return "run the rpc tail" }
func (c *tailCommand) Usage() string {
	return `tail [<flags>]:

        Monitor stdin in realtime and generate events based on
        user-specified detection rules (regexp-based). Meant to
        process third-party logs.

        The patterns file should contain one rule per line, consisting
        of a regexp bounded by / characters (Perl- style), followed by
        whitespace, followed by the event type that should be set on
        the generated event. Regular expressions must contain one
        capture group for the IP address.

        As an example, the following rule:

            /auth failure from ip ([.0-9]+)/ auth_failure

        will trigger an 'auth_failure' event with the IP extracted
        from the message.

`
}

func (c *tailCommand) SetFlags(f *flag.FlagSet) {
	f.StringVar(&c.serverAddr, "server", "", "`address` (https://host:port) of the iprep server")
	f.StringVar(&c.patternFile, "patterns", "", "`file` with patterns to load")
}

func (c *tailCommand) Execute(ctx context.Context, f *flag.FlagSet, args ...interface{}) subcommands.ExitStatus {
	if f.NArg() > 0 {
		log.Printf("error: too many arguments")
		return subcommands.ExitUsageError
	}
	if c.serverAddr == "" {
		log.Printf("error: must specify --server")
		return subcommands.ExitUsageError
	}
	if c.patternFile == "" {
		log.Printf("error: must specify --pattern")
		return subcommands.ExitUsageError
	}

	if err := c.run(ctx); err != nil {
		log.Printf("error: %v", err)
		return subcommands.ExitFailure
	}

	return subcommands.ExitSuccess
}

func (c *tailCommand) run(ctx context.Context) error {
	pf, err := os.Open(c.patternFile)
	if err != nil {
		return err
	}
	defer pf.Close()
	patterns, err := parsePatterns(pf, c.patternFile)
	if err != nil {
		return err
	}

	sub, err := submission.New(c.serverAddr, nil)
	if err != nil {
		return err
	}
	defer sub.Close()

	scanner := bufio.NewScanner(os.Stdin)
	for scanner.Scan() {
		for _, p := range patterns {
			m := p.rx.FindSubmatch(scanner.Bytes())
			if len(m) < 1 {
				continue
			}
			sub.AddEvent(&ippb.Event{
				Ip:    string(m[1]),
				Type:  p.eventType,
				Count: 1,
			})
		}
	}

	return nil
}

func init() {
	subcommands.Register(&tailCommand{}, "")
}
