package script

import (
	"log"

	"github.com/d5/tengo/compiler/token"
	"github.com/d5/tengo/objects"

	"git.autistici.org/ai3/tools/iprep/ext"
)

// Tengo object type that wraps a read-only map[string]int64, without
// requiring us to go through a map[string]interface{}.
type intMap map[string]int64

func (m intMap) String() string {
	return "<intMap>"
}

func (m intMap) TypeName() string {
	return "int-map"
}

func (m intMap) Copy() objects.Object {
	return m
}

func (m intMap) IsFalsy() bool {
	return len(m) == 0
}

func (m intMap) Equals(o objects.Object) bool {
	return false
}

func (m intMap) BinaryOp(op token.Token, rhs objects.Object) (objects.Object, error) {
	return nil, objects.ErrInvalidOperator
}

func (m intMap) IndexGet(index objects.Object) (objects.Object, error) {
	indexStr, ok := index.(*objects.String)
	if !ok {
		return nil, objects.ErrInvalidIndexType
	}
	value, ok := m[indexStr.Value]
	if !ok {
		return objects.UndefinedValue, nil
	}
	return &objects.Int{Value: value}, nil
}

func (m intMap) Iterate() objects.Iterator {
	return newIntMapIterator(m)
}

// We need to make a copy of the map to iterate on it with this API.
// Note that in the iterator, 'idx' maps to the *next* item.
type intMapIterator struct {
	arr []keyIntPair
	idx int
}

type keyIntPair struct {
	key   string
	value int64
}

func newIntMapIterator(m intMap) *intMapIterator {
	arr := make([]keyIntPair, 0, len(m))
	for k, v := range m {
		arr = append(arr, keyIntPair{key: k, value: v})
	}
	return &intMapIterator{arr: arr}
}

func (i *intMapIterator) String() string {
	return "<intMapIterator>"
}

func (i *intMapIterator) TypeName() string {
	return "int-map-iterator"
}

func (i *intMapIterator) Copy() objects.Object {
	return i
}

func (i *intMapIterator) IsFalsy() bool {
	return len(i.arr) == 0
}

func (i *intMapIterator) Equals(o objects.Object) bool {
	return false
}

func (i *intMapIterator) BinaryOp(op token.Token, rhs objects.Object) (objects.Object, error) {
	return nil, objects.ErrInvalidOperator
}

func (i *intMapIterator) Next() bool {
	i.idx++
	return i.idx <= len(i.arr)
}

func (i *intMapIterator) Key() objects.Object {
	return &objects.String{Value: i.arr[i.idx-1].key}
}

func (i *intMapIterator) Value() objects.Object {
	return &objects.Int{Value: i.arr[i.idx-1].value}
}

// A Tengo callable to look up data in external sources.
type extLookupFunc map[string]ext.ExternalSource

func (f extLookupFunc) String() string {
	return "<extLookupFunc>"
}

func (f extLookupFunc) TypeName() string {
	return "ext-lookup-func"
}

func (f extLookupFunc) Copy() objects.Object {
	return f
}

func (f extLookupFunc) IsFalsy() bool {
	return len(f) == 0
}

func (f extLookupFunc) Equals(o objects.Object) bool {
	return false
}

func (f extLookupFunc) BinaryOp(op token.Token, rhs objects.Object) (objects.Object, error) {
	return nil, objects.ErrInvalidOperator
}

func (f extLookupFunc) Call(args ...objects.Object) (ret objects.Object, err error) {
	// Expected arguments: source name, IP.
	if len(args) != 2 {
		return nil, objects.ErrWrongNumArguments
	}

	name, ok := objects.ToString(args[0])
	if !ok {
		return nil, objects.ErrInvalidArgumentType{
			Name:     "source_name",
			Expected: "string",
			Found:    args[0].TypeName(),
		}
	}

	ip, ok := objects.ToString(args[1])
	if !ok {
		return nil, objects.ErrInvalidArgumentType{
			Name:     "ip",
			Expected: "string",
			Found:    args[1].TypeName(),
		}
	}

	// Invoke the external source lookup method.
	if f == nil {
		return nil, objects.ErrInvalidIndexValueType
	}
	src, ok := f[name]
	if !ok {
		return nil, objects.ErrInvalidIndexValueType
	}
	result, err := src.LookupIP(ip)
	if err != nil {
		log.Printf("external source lookup error: ip=%s: %v", ip, err)
		return objects.UndefinedValue, nil
	}
	return result, nil
}
